@@ -1,5 +1,5 @@
 r"""
-Patterns and functions to detect arXiv ids and Urls in text.
+Patterns and functions to detect arXiv ids and URLs in text.
 
 Functions to detect arXiv ids, URLs and DOI in text.
 Functions to transform them to <a> tags.
@@ -19,8 +19,8 @@ HTTP URLs:      http://something.org/myPaper/1234.12345
 FTP URLs:       ftp://example.com/files/1234.12345
 
 Just matching for arXiv ids with \d{4}\.\d{4,5} will match several of
-these. To deal with this we are priortizing the matches and
-interupting once one is found.
+these. To deal with this we are prioritizing the matches and
+interrupting once one is found.
 
 We should probably match DOIs first because they are the source of a
 lot of false positives for arxiv matches.
@@ -33,21 +33,21 @@ handle FTP addresses.
 Updated 20 March, 2019: refactored to provide independent bleach attribute
 callbacks for each kind of link.
 """
-from typing import Optional, List, Pattern, Match, Tuple, Callable, \
-    NamedTuple, Dict, Union, Mapping
+from typing import List, Pattern, Tuple, Callable, Dict, Union, Text, Any, \
+    MutableMapping
 import re
-from functools import partial, reduce
+from functools import reduce
 from urllib.parse import quote, urlparse
 
-from flask import url_for
-from jinja2 import Markup, escape
+from flask import url_for, g
 import bleach
 
+from arxiv.taxonomy import CATEGORIES
 from arxiv import identifier
 from . import clickthrough
 
 
-Attrs = Dict[Union[str, Tuple[None, str]], str]
+Attrs = MutableMapping[Any, Text]
 Callback = Callable[[Attrs, bool], Attrs]
 Callable_Linker = Callable[[str], str]
 
@@ -65,18 +65,19 @@ DOI = re.compile(   # '10.1145/0001234.1234567'
 Pattern for matching DOIs.
 
 We should probably match DOIs first because they are the source of a
-lot of false positives for arxiv matches.
+lot of false positives for arxiv id matches.
 
 Only using the most general expression from
 https://www.crossref.org/blog/dois-and-matching-regular-expressions/
 """
 
 
-BROAD_DOI=re.compile(r'(?P<doi>10\.\d{4,5}\/\S+)', re.I)
+BROAD_DOI = re.compile(r'(?P<doi>\d{2,3}.\d{4,5}\/\S+)', re.I)
 """
 Very broad pattern for matching DOIs in the abs DOI field.
 
 Ex. 10.1175/1520-0469(1996)053<0946:ASTFHH>2.0.CO;2
+Ex. 21.11130/00-1735-0000-0005-146A-E
 """
 
 ARXIV_PATTERNS = [
@@ -91,7 +92,7 @@ ARXIV_ID = re.compile(
 )
 
 OKCHARS = r'([a-z0-9,_.\-+~:]|%[a-f0-9]*)'
-"""Chacters that are acceptable during PATH, QUERY and ANCHOR parts"""
+"""Characters that are acceptable during PATH, QUERY and ANCHOR parts"""
 
 PATH = rf'(?P<PATH>(/{OKCHARS}*)+)?'
 """Regex for path part of URLs for use in urlize"""
@@ -146,30 +147,34 @@ def _extend_class_attr(attrs: Attrs, new_class: str) -> Attrs:
 
 
 def _add_rel(attrs: Attrs, new: bool = False) -> Attrs:
-    """Check if href is interal or external and adds attrs as appropriate."""
+    """Check if href is internal or external and adds attrs as appropriate."""
     o = urlparse(attrs[(None, 'href')])
     if not o.netloc.split(':')[0].endswith('arxiv.org'):   # External link?
         return _add_rel_external(attrs, new)
     else:
         return _add_rel_internal(attrs, new)
 
+
 def _add_rel_external(attrs: Attrs, new: bool = False) -> Attrs:
-    """Always adds external rel."""
+    """Add an external rel."""
     # noopener for security reasons
-    # nofollow to disencentivie arXiv articles for SEO
+    # nofollow to disincentivize arXiv articles for SEO
     # external says that link is away from arxiv
     attrs[(None, 'rel')] = 'external noopener nofollow'
     _extend_class_attr(attrs, 'link-external')
     return attrs
 
+
 def _add_rel_internal(attrs: Attrs, new: bool = False) -> Attrs:
-    """Always adds interal rel."""
+    """Add an internal rel."""
     _extend_class_attr(attrs, 'link-internal')
     return attrs
 
+
 def _this_url_text(attrs: Attrs, new: bool = False) -> Attrs:
     o = urlparse(attrs[(None, 'href')])
-    attrs['_text'] = f'this {o.scheme} URL'    # Replaces the link text.
+    if o.hostname and not o.hostname.endswith('arxiv.org'):
+        attrs['_text'] = f'this {o.scheme} URL'    # Replaces the link text.
     return attrs
 
 
@@ -231,6 +236,42 @@ def _handle_doi_url(attrs: Attrs, new: bool = False) -> Attrs:
     return attrs
 
 
+def _handle_broad_doi_url(attrs: Attrs, new: bool = False) -> Attrs:
+    """
+    Handle doi from DOI field.
+
+    It is always just a DOI so turn it into a DOI link.
+    """
+    doi_attrs = _handle_doi_url(attrs, new)
+    if (None, 'data-doi') in doi_attrs:
+        return doi_attrs
+    else:
+        target = attrs['_text']
+        attrs[(None, 'href')] = clickthrough_url_for_doi(target)
+        attrs[(None, 'data-doi')] = target
+        return attrs
+
+
+ENDS_WITH_TLD = r".*\.(" + TLDS + ")$"
+CATEGORIES_THAT_COULD_BE_HOSTNAMES = '|'.join([cat_id for cat_id in CATEGORIES.keys()
+                                               if re.search(ENDS_WITH_TLD, cat_id, re.IGNORECASE)])
+DONT_URLIZE_CATS = re.compile(CATEGORIES_THAT_COULD_BE_HOSTNAMES,
+                              re.IGNORECASE | re.UNICODE)
+
+
+def _dont_urlize_arxiv_categories(attrs: Attrs, new: bool = False) -> Attrs:
+    """
+    Prevent urlizing archive categories that look like hostnames.
+
+    Ex. don't urlize math.CO but do urlize supermath.co
+    """
+    url = urlparse(attrs[(None, 'href')])
+    if DONT_URLIZE_CATS.match(url.netloc):
+        return None  # type: ignore
+    else:
+        return attrs
+
+
 PATTERNS = {
     'doi': DOI,
     'doi_field': BROAD_DOI,
@@ -246,9 +287,9 @@ for just the abs DOI field."""
 
 callbacks = {
     'doi':      [_handle_doi_url, _add_scheme_info, _add_rel_external],
-    'doi_field': [_handle_doi_url, _add_scheme_info, _add_rel_external],
+    'doi_field': [_handle_broad_doi_url, _add_scheme_info, _add_rel_external],
     'arxiv_id': [_handle_arxiv_url, _add_scheme_info],
-    'url':      [_this_url_text, _add_rel, _add_scheme_info ]
+    'url':      [_dont_urlize_arxiv_categories, _this_url_text, _add_rel, _add_scheme_info]
 }
 """Bleach attribute callbacks for each kind."""
 
@@ -256,15 +297,14 @@ callbacks = {
 def _get_pattern(kinds: List[str]) -> Pattern:
     return re.compile(
         '|'.join([rf'(?:{PATTERNS[kind].pattern})' for kind
-                 in sorted(kinds, key=lambda kind: ORDER.index(kind))]),
+                  in sorted(kinds, key=lambda kind: ORDER.index(kind))]),
         re.IGNORECASE | re.VERBOSE | re.UNICODE
     )
 
-
 def _get_linker_of_kind(kind: str) -> Callable_Linker:
     # The returned object will attempt to match url_re in the input string.
     # For each url_re that matches, it will run all of the attribute
-    # adjsuting callbacks and then stick the tag in the output string
+    # adjusting callbacks and then stick the tag in the output string
     # replacing the matched text.
     linker: Callable[[str], str] = bleach.linkifier.Linker(
         callbacks=callbacks[kind],
@@ -274,16 +314,44 @@ def _get_linker_of_kind(kind: str) -> Callable_Linker:
 
 
 def _compose_list_of_funcs(fv: List[Callable_Linker]) -> Callable_Linker:
-    """Returns function that calls fv functions one after another."""
+    """Return function that calls fv functions one after another."""
     if not fv:
         return lambda x: x
     return reduce(lambda f, g: lambda x: f(g(x)), fv[1:], fv[0])
 
 
+def _deferred_thread_local_linker_of_kind(kind: str) -> Callable_Linker:
+    # Bleach linkifiers are not thread safe. This puts them on Flask.g.
+    #
+    # This will return a Callable that creates the bleach linkifier
+    # only when called with the string to linkify. This is so the
+    # bleach linkifier can be put into a Flask.g of each
+    # individual WSGI thread.
+
+    def deferred(instr: str) -> str:
+        # This must not be called while the app is starting up, only
+        # when the linkiers are being used in requests. So the
+        # function is created/cached at call time.
+        if 'linkers' not in g:
+            g.linkers = {}
+        if kind not in g.linkers:
+            g.linkers[kind] = _get_linker_of_kind(kind)
+
+        return g.linkers[kind](instr) #type: ignore
+
+    return deferred
+
+
 def _get_linker(kinds: List[str]) -> Callable_Linker:
+    # Gets composed, thread-local, deferred linkers.
+    # This is intended to be run as the flask app starts up
+    # or in the loading of packages.
+    # The actual creating of thread local bleach objects is
+    # done in _deferred_thread_local_linker_of_kind.
     if len(kinds) > 1 and 'doi_field' in kinds:
-        raise ValueError('doi_field should not be used in combination with other kinds')
-    ordered_linkers = [_get_linker_of_kind(kind) for kind
+        raise ValueError(
+            'doi_field should not be used in combination with other kinds')
+    ordered_linkers = [_deferred_thread_local_linker_of_kind(kind) for kind
                        in sorted(kinds, key=lambda kind: ORDER.index(kind))]
     return _compose_list_of_funcs(ordered_linkers)
 
